/**
 * Copyright <2022> <Innovative Sensor Technology IST AG>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 * 
 * 4. This license is granted under the laws of Switzerland, and the Courts of
 * Basle City shall have exclusive jurisdiction for any dispute arising in the
 * context of this license.  
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (EXCLUDING GROSS NEGLIGENCE AND INTENT
 * AS WELL AS OTHER MANDATORY LEGAL PROVISONS SUCH AS PRODUCT LIABILITY LAWS)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE.
 */


//Projekt:		Software I2C-Address
//Datei:		main.c
//Datum:		2012-03-28
//Autor:		Mike Steitz
//
// Software I2C-Master to change the I2C-Address of a HYT-sensor of the company IST AG
// Program is tested on a ATMEL AtTiny 84 and the IDE AVR-Studio Version 4.18 Build 716
// the sensor is connected to ports PA2 (SCL) and PA3 (SDA)
// The VDD pin of the sensor is connected to port PA0
// Result of operation will be signaled on port PA4 (success) and PA5 (error)


#include <avr/io.h>
#include <avr/interrupt.h>


#include "timer.h"
#include "sw_master.h"

#define I2C_READY_         	0x01
#define I2C_BUSY_		  	0x02

// port definition to power the sensor and signal the results
#define DO_DDR	       		DDRA
#define DO_PORT      		PORTA
#define DO_POWER_MASK	   	(1<<DDA0)
#define DO_READY_MASK		(1<<DDA4)
#define DO_ERROR_MASK		(1<<DDA5)

// new address to be written to the sensor
#define I2C_NEW_ADDRESS		0x29

typedef enum {
	POWER_OFF_ON			= 0x00,
	SET_COMMAND_MODE		= 0x01,
	WRITE_EEPROM_ADDRESS	= 0x02,
	READ_I2C_ADDRESS		= 0x03,
	WRITE_I2C_ADDRESS		= 0x04,
	READ_RESPONSE			= 0x05,
	ADR_CHANGE_READY		= 0x06,
} adr_change_state_t;

adr_change_state_t 	adr_change_state;


// Init the ports for power-supply and signaling as output
static void init_do(void) {

	DO_DDR |= DO_POWER_MASK;
	DO_DDR |= DO_READY_MASK;
	DO_DDR |= DO_ERROR_MASK;

	return;
}

// function to control the power supply of the sensor
static void do_power_hyt(unsigned char on) {

	if (on) {
		DO_PORT |= (DO_POWER_MASK);
	} else {
		DO_PORT &= ~(DO_POWER_MASK);
	}
	return;
}

// function to control the "ready"-output
static void do_ready_hyt(unsigned char on) {

	if (on) {
		DO_PORT |= (DO_READY_MASK);
	} else {
		DO_PORT &= ~(DO_READY_MASK);
	}
	return;
}

// function to control the "error"-output
static void do_error_hyt(unsigned char on) {

	if (on) {
		DO_PORT |= (DO_ERROR_MASK);
	} else {
		DO_PORT &= ~(DO_ERROR_MASK);
	}
	return;
}



int main(void) {
	uint8_t   			msg[3];
	uint8_t   			rval;
	uint8_t   			slaveAddress; 
	uint8_t   			new_slaveAddress; 
	uint8_t   			i2c_status;
	uint8_t   			adr_low = 0;
	uint8_t   			adr_high = 0;
	adr_change_state_t 	next_state = POWER_OFF_ON;
	i2c_errorstate_t 	i2c_errorstate_copy;
	uint8_t   			found;
	uint8_t   			found_copy;
	uint8_t   			i;



	cli(); // disable interrupt

	// Init Timer and digital outputs
	timer1_init();
	init_do();

	do_power_hyt(1);
	do_ready_hyt(0);
	do_error_hyt(0);


	// Init I2C-Master
	rval = init_i2c_master();
	if (rval == I2C_OK) {
		i2c_status = I2C_READY_;
	} else {
		i2c_status = I2C_BUSY_;
	}


	sei(); // enable_interrupt

	// detect the I2C-address of the sensor
	found = 0;
	for (i=0; i<=0x7F; i++) {
  		rval = i2c_master_write(i, &msg[0], 3); 
		if (rval == I2C_OK) {
			slaveAddress = i;
			found = 1;
			break;
		} else {}
	}


	if (found == 1) {
	  	adr_change_state = POWER_OFF_ON;
		new_slaveAddress = I2C_NEW_ADDRESS;

	
		while (i2c_status == I2C_READY_ && adr_change_state < ADR_CHANGE_READY) {
	    	switch (adr_change_state) {
  				// toggle power supply of the sensor (power-off -> power-on)
				case POWER_OFF_ON:
  					do_power_hyt(0);
					// delay 100ms
					i2c_time_out_timer = 0;
					while (i2c_time_out_timer < 10) {
					}
  					do_power_hyt(1);
					adr_change_state = SET_COMMAND_MODE;
					break;
				// send message to enter the command-mode
				case SET_COMMAND_MODE:
  					msg[0] = 0xA0;
  					msg[1] = 0;
  					msg[2] = 0;
	  				rval = i2c_master_write(slaveAddress, &msg[0], 3); 
  					if (rval == I2C_OK) {
						adr_change_state = READ_RESPONSE;
						next_state = WRITE_EEPROM_ADDRESS;
					} else {
    					i2c_status = I2C_BUSY_;
					}
					break;
				// write the EEPROM-address that contains the I2C-address
				case WRITE_EEPROM_ADDRESS:
        			msg[0] = 0x1C;
	        		msg[1] = 0;
    	    		msg[2] = 0;
        			rval = i2c_master_write(slaveAddress, &msg[0], 3); 
        			if (rval == I2C_OK) {
						adr_change_state = READ_RESPONSE;
						next_state = READ_I2C_ADDRESS;
					} else {
    					i2c_status = I2C_BUSY_;
					}
					break;
				// read the I2C-address
				case READ_I2C_ADDRESS:
        	  		rval = i2c_master_read(slaveAddress, &msg[0], 3);
	      			if (rval == I2C_OK && msg[0] == 0x81) {
						adr_high = msg[1];
						adr_low = msg[2];
						adr_change_state = WRITE_I2C_ADDRESS;
					} else {
    					i2c_status = I2C_BUSY_;
					}
					break;
				// write the new I2C-address
				case WRITE_I2C_ADDRESS:
				  	msg[0] = 0x5C;
					msg[1] = adr_high;
					adr_low &= 0x80;
					adr_low |= (new_slaveAddress & 0x7F);
				  	msg[2] = adr_low;
        	      	rval = i2c_master_write(slaveAddress, &msg[0], 3); 
	        	  	if (rval == I2C_OK) {
						adr_change_state = READ_RESPONSE;
						next_state = ADR_CHANGE_READY;
					} else {
    					i2c_status = I2C_BUSY_;
					}
					break;
				case ADR_CHANGE_READY:
					break;
				// check the answer if there was a write access
				case READ_RESPONSE:
    				// 100ms delay (response-time)
					i2c_time_out_timer = 0;
    				while (i2c_time_out_timer < 10) {
    				}
					// read the status
    				rval = i2c_master_read(slaveAddress, &msg[0], 3);
					if (rval == I2C_OK && msg[0] == 0x81) {
						adr_change_state = next_state;
					} else {
	    				i2c_status = I2C_BUSY_;
					}
					break;
			}
		}
	} else {}

	// set the output to signal an error and the ready-state
	if (i2c_status == I2C_BUSY_ || found == 0) {
		do_error_hyt(1);
    } else {}
	do_ready_hyt(1);

	// place your own error-handling here
	while (1) {
		i2c_errorstate_copy = i2c_errorstate;
		found_copy = found;
	}
   
}



